@@ -37,9 +37,6 @@
 #include "amdgpu.h"
 
 typedef struct {
-	unsigned fb_id;
-	drmModeFBPtr mode_fb;
-	int cpp;
 	ScrnInfoPtr scrn;
 #ifdef HAVE_LIBUDEV
 	struct udev_monitor *uevent_monitor;
@@ -55,10 +52,9 @@ typedef struct {
 } drmmode_rec, *drmmode_ptr;
 
 typedef struct {
-	int fd;
-	unsigned old_fb_id;
-	int flip_count;
+	struct drmmode_fb *fb;
 	void *event_data;
+	int flip_count;
 	unsigned int fe_frame;
 	uint64_t fe_usec;
 	xf86CrtcPtr fe_crtc;
@@ -66,10 +62,14 @@ typedef struct {
 	amdgpu_drm_abort_proc abort;
 } drmmode_flipdata_rec, *drmmode_flipdata_ptr;
 
+struct drmmode_fb {
+	int refcnt;
+	uint32_t handle;
+};
+
 struct drmmode_scanout {
 	struct amdgpu_buffer *bo;
 	PixmapPtr pixmap;
-	unsigned fb_id;
 	int width, height;
 };
 
@@ -80,15 +80,16 @@ typedef struct {
 	struct amdgpu_buffer *cursor_buffer;
 	struct drmmode_scanout rotate;
 	struct drmmode_scanout scanout[2];
-	struct drmmode_scanout scanout_destroy[2];
 	DamagePtr scanout_damage;
+	Bool ignore_damage;
 	RegionRec scanout_last_region;
 	unsigned scanout_id;
 	Bool scanout_update_pending;
 	Bool tear_free;
+
+	PixmapPtr prime_scanout_pixmap;
+
 	int dpms_mode;
-	/* For when a flip is pending when DPMS off requested */
-	int pending_dpms_mode;
 	CARD64 dpms_last_ust;
 	uint32_t dpms_last_seq;
 	int dpms_last_fps;
@@ -96,8 +97,18 @@ typedef struct {
 
 	/* Modeset needed for DPMS on */
 	Bool need_modeset;
-	/* A flip is pending for this CRTC */
-	Bool flip_pending;
+	/* A flip to this FB is pending for this CRTC */
+	struct drmmode_fb *flip_pending;
+	/* The FB currently being scanned out by this CRTC, if any */
+	struct drmmode_fb *fb;
+
+#ifdef HAVE_PRESENT_H
+	/* Deferred processing of Present vblank event */
+	uint64_t present_vblank_event_id;
+	uint64_t present_vblank_usec;
+	unsigned present_vblank_msc;
+	Bool present_flip_expected;
+#endif
 } drmmode_crtc_private_rec, *drmmode_crtc_private_ptr;
 
 typedef struct {
@@ -128,6 +139,52 @@ enum drmmode_flip_sync {
 };
 
 
+/* Can the page flip ioctl be used for this CRTC? */
+static inline Bool
+drmmode_crtc_can_flip(xf86CrtcPtr crtc)
+{
+	drmmode_crtc_private_ptr drmmode_crtc = crtc->driver_private;
+
+	return crtc->enabled &&
+		drmmode_crtc->dpms_mode == DPMSModeOn &&
+		!drmmode_crtc->rotate.bo &&
+		(drmmode_crtc->tear_free ||
+		 !drmmode_crtc->scanout[drmmode_crtc->scanout_id].bo);
+}
+
+
+static inline void
+drmmode_fb_reference_loc(int drm_fd, struct drmmode_fb **old, struct drmmode_fb *new,
+			 const char *caller, unsigned line)
+{
+	if (new) {
+		if (new->refcnt <= 0) {
+			FatalError("New FB's refcnt was %d at %s:%u",
+				   new->refcnt, caller, line);
+		}
+
+		new->refcnt++;
+	}
+
+	if (*old) {
+		if ((*old)->refcnt <= 0) {
+			FatalError("Old FB's refcnt was %d at %s:%u",
+				   (*old)->refcnt, caller, line);
+		}
+
+		if (--(*old)->refcnt == 0) {
+			drmModeRmFB(drm_fd, (*old)->handle);
+			free(*old);
+		}
+	}
+
+	*old = new;
+}
+
+#define drmmode_fb_reference(fd, old, new) \
+	drmmode_fb_reference_loc(fd, old, new, __func__, __LINE__)
+
+
 extern int drmmode_page_flip_target_absolute(AMDGPUEntPtr pAMDGPUEnt,
 					     drmmode_crtc_private_ptr drmmode_crtc,
 					     int fb_id, uint32_t flags,
@@ -149,21 +206,33 @@ extern Bool drmmode_set_desired_modes(ScrnInfoPtr pScrn, drmmode_ptr drmmode,
 extern void drmmode_copy_fb(ScrnInfoPtr pScrn, drmmode_ptr drmmode);
 extern Bool drmmode_setup_colormap(ScreenPtr pScreen, ScrnInfoPtr pScrn);
 
-extern void drmmode_scanout_free(ScrnInfoPtr scrn);
+extern void drmmode_crtc_scanout_destroy(drmmode_ptr drmmode,
+					 struct drmmode_scanout *scanout);
+void drmmode_crtc_scanout_free(drmmode_crtc_private_ptr drmmode_crtc);
+PixmapPtr drmmode_crtc_scanout_create(xf86CrtcPtr crtc,
+				      struct drmmode_scanout *scanout,
+				      int width, int height);
 
 extern void drmmode_uevent_init(ScrnInfoPtr scrn, drmmode_ptr drmmode);
 extern void drmmode_uevent_fini(ScrnInfoPtr scrn, drmmode_ptr drmmode);
 
+Bool drmmode_set_mode(xf86CrtcPtr crtc, struct drmmode_fb *fb,
+		      DisplayModePtr mode, int x, int y);
+
 extern int drmmode_get_crtc_id(xf86CrtcPtr crtc);
 extern int drmmode_get_pitch_align(ScrnInfoPtr scrn, int bpe);
-extern void drmmode_clear_pending_flip(xf86CrtcPtr crtc);
 Bool amdgpu_do_pageflip(ScrnInfoPtr scrn, ClientPtr client,
 			PixmapPtr new_front, uint64_t id, void *data,
-			int ref_crtc_hw_id, amdgpu_drm_handler_proc handler,
+			xf86CrtcPtr ref_crtc, amdgpu_drm_handler_proc handler,
 			amdgpu_drm_abort_proc abort,
 			enum drmmode_flip_sync flip_sync,
 			uint32_t target_msc);
 int drmmode_crtc_get_ust_msc(xf86CrtcPtr crtc, CARD64 *ust, CARD64 *msc);
 int drmmode_get_current_ust(int drm_fd, CARD64 * ust);
 
+Bool drmmode_wait_vblank(xf86CrtcPtr crtc, drmVBlankSeqType type,
+			 uint32_t target_seq, unsigned long signal,
+			 uint64_t *ust, uint32_t *result_seq);
+
+
 #endif
